unit ReadBDEIni;

interface

uses
 Windows, SysUtils, Classes, ComCtrls, StdCtrls, IniFiles;

type
  ESetupBDE = Exception;
  TSetupBDE = class
  private
    FTempFile, FIniFile: string;
    FTV: TTreeView;
    function GetIdentify(Section: string): string;
    function GetValue(Section: string): string;
    procedure OpenIniFile(FileName: string);
    procedure Win32Chk(Win32OK: Boolean);
  public
    IniFIleChanged, BDEConfigOnly: boolean;
    constructor Create(FileName: string; TreeView: TTreeView);
    procedure Refresh;
    procedure View(TreeView: TTreeView);
    destructor Destroy; override;
    function GetIniFileName: string;
    function GetTempFileName: string;
    function GetAsString(Section, Ident: string): string;
    function Update(Node: TTreeNode; Value: string): boolean;
    procedure SaveIniFile;
    procedure SaveIniFileAs(NewName: string);
    procedure RegisterBDE(CreateBorlandKey: boolean);
    procedure TestBDE; virtual;    
  end;

implementation

uses BDE, Dialogs, Registry;

const
  RootNode: TTreeNode = nil;
  cWorkName = '~working.ini';


constructor TSetupBDE.Create(FileName: string; TreeView: TTreeView);
begin
  BDEConfigOnly := False;
  IniFileChanged := False;
  View(TreeView);
  OpenIniFile(FileName);
end;

destructor TSetupBDE.Destroy;
begin
  DeleteFile(FTempFile);
  inherited;
end;

procedure TSetupBDE.Win32Chk(Win32OK: Boolean);
const
  ERRORSIZE = 300;  // Maximum error string size
var
  Buffer: String;

begin
  if Win32OK = False then
  begin
    SetLength(Buffer, ERRORSIZE);
    if FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,  nil, GetLastError, 0, PChar(Buffer),
              ERRORSIZE, nil) = 0 then
      Buffer := 'Internal Error: Error creating Error String';
    SetLength(Buffer, StrLen(PChar(Buffer)));
    raise Exception.Create(Buffer);
  end;
end;


procedure TSetupBDE.Refresh;
var
  Test: string;
  InIni: TIniFile;
  Sections, SectionItems: TStringList;
  SectionCount, ItemCount: word;
  ParentNode: TTreeNode;

begin
  if FTV <> nil then
  begin
    InIni := TIniFile.Create(FIniFile);
    Sections := TStringList.Create;
    try
      if BDEConfigOnly = True then
      begin
        test := InIni.ReadString('Database Engine', 'DLLPATH', 'OOPS');
        if test = 'OOPS' then
          raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
                   #10#13'   FILE: ' + FIniFile);
        test := InIni.ReadString('Database Engine', 'CONFIGFILE01', 'OOPS');
        if test = 'OOPS' then
          raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
                   #10#13'   FILE: ' + FIniFile);
        test := InIni.ReadString('BLW32', 'BLAPIPATH', 'OOPS');
        if test = 'OOPS' then
          raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
                   #10#13'   FILE: ' + FIniFile);
        test := InIni.ReadString('Database Engine\Settings\SYSTEM\INIT',
                        'MAXFILEHANDLES', 'OOPS');
        if test = 'OOPS' then
          raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
                   #10#13'   FILE: ' + FIniFile);
      end;
      FTV.Items.Clear;
      InIni.ReadSections(Sections);
      for SectionCount := 0 to Sections.Count - 1 do
      begin
        SectionItems := TStringList.Create;
        try
          ParentNode := FTV.Items.Add(RootNode, Sections.Strings[SectionCount]);
          InIni.ReadSectionValues(Sections.Strings[SectionCount], SectionItems);
          for ItemCount := 0 to SectionItems.Count - 1 do
            FTV.Items.AddChild(ParentNode,
                GetIdentify(SectionItems.Strings[ItemCount]));
        finally
          SectionItems.Free;
        end;
      end;
    finally
      InIni.Free;
      Sections.Free;
    end;
  end;
end;

procedure TSetupBDE.OpenIniFile(FileName: string);
begin
  FIniFile := FileName;
  FTempFile := ExtractFilePath(FileName) + cWorkName;
  Win32Chk(CopyFile(Pchar(FileName), Pchar(FTempFile), FALSE));
  Win32Chk(SetFileAttributes(PChar(FTempFile), FILE_ATTRIBUTE_NORMAL));
end;

procedure TSetupBDE.View(TreeView: TTreeView);
begin
  FTV := TreeView;
end;

function TSetupBDE.GetIdentify(Section: string): string;
begin
  if Pos('=', Section) = 0 then
    raise ESetupBDE.Create('Could not find identifier terminator (=)');

  Result := Copy(Section, 1, Pos('=', Section) - 1);

end;

function TSetupBDE.GetIniFileName: string;
begin
  Result := FIniFile;
end;

function TSetupBDE.GetTempFileName: string;
begin
  Result := FTempFile;
end;

function TSetupBDE.GetAsString(Section, Ident: string): string;
var
  InIni: TIniFile;

begin
  InIni := TIniFile.Create(FTempFile);
  try
    Result := InIni.ReadString(Section, Ident, 'ERROR');
    if Result = 'ERROR' then
      raise ESetupBDE.Create('Could not find identifier in INI file');
  finally
    InIni.Free;
  end;
end;

function TSetupBDE.Update(Node: TTreeNode; Value: string): boolean;
var
  OutputIni: TIniFile;
  OldValue: string;

begin
  Result := False;
  if Node.Level <> 1 then
    raise ESetupBDE.Create('Updates are only allowed on values');
  OutputIni := TIniFile.Create(FTempFile);
  try
    OldValue := OutputIni.ReadString(Node.Parent.Text, Node.Text, 'OOPS');
    if OldValue = 'OOPS' then
      raise ESetupBDE.Create('Could not find item to update in INI file.' +
              #10#13'  FILE: ' + FTempFile);
    if CompareStr (OldValue, Value) <> 0 then
    begin
      OutputIni.WriteString(Node.Parent.Text, Node.Text, Value);
      IniFileChanged := True;
      Result := True;
    end;
  finally
    OutputIni.Free;
  end;
end;

procedure TSetupBDE.SaveIniFile;
begin
  Win32Chk(CopyFile(PChar(FTempFile), PChar(FIniFile), FALSE));
  IniFileChanged := False;  
end;

procedure TSetupBDE.SaveIniFileAs(NewName: string);
begin
  Win32Chk(CopyFile(PChar(FTempFile), PChar(NewName), FALSE));
  FIniFile := NewName;
  IniFileChanged := False;
  Refresh;
end;


procedure CheckBDE(rslt: DBIResult);
var
  ErrInfo: DBIErrInfo;
  Result: string;

begin
  if rslt <> DBIERR_NONE then
  begin
    Result := '';
    FillChar(ErrInfo, Sizeof(ErrInfo), #0);
    DbiGetErrorInfo(False, ErrInfo);
    // There is information in the ErrInfo structure...
    if ErrInfo.iError = rslt then
    begin
      if StrLen(ErrInfo.szContext[1]) > 0 then
        Result := Result + ErrInfo.szContext[1] + #10#13;
      if StrLen(ErrInfo.szContext[2]) > 0 then
        Result := Result + ErrInfo.szContext[2] + #10#13;
      if StrLen(ErrInfo.szContext[3]) > 0 then
        Result := Result + ErrInfo.szContext[3] + #10#13;
      if StrLen(ErrInfo.szContext[4]) > 0 then
        Result := Result + ErrInfo.szContext[4] + #10#13;
    end
    else
    // Just get the error with DbiGetErrorString...
    begin
      SetLength(Result, DBIMAXMSGLEN + 1);
      DbiGetErrorString(rslt, PChar(Result));
      SetLength(Result, DBIMAXMSGLEN + 1);
    end;
    raise Exception.Create(Result);
  end;
end;

procedure TSetupBDE.TestBDE;
var
  hSes: hDBISes;
  hCur: hDBICur;

begin
  CheckBDE(DbiInit(nil));
  try try
    CheckBDE(DbiStartSession('TestSession', hSes, nil));
    CheckBDE(DbiCloseSession(hSes));
    CheckBDE(DbiOpenCfgInfoList(nil, dbiREADWRITE, cfgPersistent, '\', hCur));
    CheckBDE(DbiCloseCursor(hCur));
    CheckBDE(DbiAddPassword('HELLO'));
    CheckBDE(DbiDropPassword('HELLO'));
    MessageDlg('All tests completed successfully', mtInformation, [mbOk], 0);
  except
     raise Exception.Create('An error occured while testing BDE initialization.' +
            #10#13'  ' + Exception(ExceptObject).Message);
  end;
  finally
    DbiExit;
  end;
end;

function TSetupBDE.GetValue(Section: string): string;
begin
  if Pos('=', Section) = 0 then
    raise ESetupBDE.Create('Could not find identifier terminator (=)');

  Result := Copy(Section, Pos('=', Section) + 1,
                  Length(section) - (Pos('=', Section) - 1));
end;

procedure TSetupBDE.RegisterBDE(CreateBorlandKey: boolean);
const
 SoftwareKey = 'SOFTWARE';
 BorlandKey = 'SOFTWARE\Borland\';

var
  Reg: TRegistry;
  Ini: TIniFile;
  Sections, SectionItems: TStringList;
  test: string;
  i, w: word;

begin
  Reg := nil; Ini := nil; Sections := nil; SectionItems := nil;
  try
    Reg := TRegistry.Create;
    Ini := TIniFile.Create(FIniFile);
    Sections := TStringList.Create;
    SectionItems := TStringList.Create;

    Reg.RootKey := HKEY_LOCAL_MACHINE;
    Reg.LazyWrite := False;
    if Reg.KeyExists(SoftwareKey) = False then
      raise Exception.Create(Softwarekey + ' key does not exist.');
    if Reg.KeyExists(BorlandKey) = False then
      if CreateBorlandKey = False then
        raise Exception.Create(BorlandKey + ' key does not exist.')
      else
        Reg.Createkey(BorlandKey);
    test := Ini.ReadString('Database Engine', 'DLLPATH', 'OOPS');
    if test = 'OOPS' then
      raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
               #10#13'   FILE: ' + FIniFile);
    test := Ini.ReadString('Database Engine', 'CONFIGFILE01', 'OOPS');
    if test = 'OOPS' then
      raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
               #10#13'   FILE: ' + FIniFile);
    test := Ini.ReadString('BLW32', 'BLAPIPATH', 'OOPS');
    if test = 'OOPS' then
      raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
               #10#13'   FILE: ' + FIniFile);
    test := Ini.ReadString('Database Engine\Settings\SYSTEM\INIT',
                    'MAXFILEHANDLES', 'OOPS');
    if test = 'OOPS' then
      raise ESetupBDE.Create('Could not find BDE entries in open INI file.' +
               #10#13'   FILE: ' + FIniFile);
    Ini.ReadSections(Sections);
    for w := 0 to Sections.Count - 1 do
    begin
      Reg.OpenKey(BorlandKey + Sections.Strings[w] + '\', True);
      SectionItems.Clear;
      Ini.ReadSectionValues(Sections.Strings[w], SectionItems);
      for i := 0 to SectionItems.Count - 1 do
      begin
        Reg.WriteString(GetIdentify(SectionItems.Strings[i]),
          GetValue(SectionItems.Strings[i]));
      end;
      Reg.CloseKey;
    end;
  finally
    if Reg <> nil then
      Reg.Free;
    if Ini <> nil then
      Ini.Free;
    if Sections <> nil then
      Sections.Free;
    if SectionItems <> nil then
      SectionItems.Free;
  end;
end;

end.
